import gym
from gym import spaces

from ..render.renderer import Renderer 
from ..physics.world import Body
from ..physics.world import World
from ..physics.box_body import BoxBody
from ..physics.circle_body import CircleBody

import random
import math
import numpy as np
from typing import Iterator, Tuple

class PodWorldEnv(gym.Env):
    metadata = {'render.modes': ['human', 'rgb_array'],
                'video.frames_per_second': 60} #TODO: support 'ansi'

    OBJ_TYPE_ROCK = 1
    OBJ_TYPE_FOOD = 2
    OBJ_TYPE_BAR = 3
    OBJ_TYPE_WALL = 4
    OBJ_TYPE_AGENT = 0

    EATEN_FOOD_COLOR = (194, 199, 190, 255)
    AVAIL_FOOD_COLOR = (74, 198, 73, 255)
    OBS_COLOR = (162, 110, 180,255)


    def __init__(self, food_count=10, obs_count=30, xmax=2560, ymax=1440, seed=42,
        box_size=(40.0, 40.0), circle_radius=30.0, box_mass=1.0, circle_mass=1.0,
        bar_count=4, bar_size=(900.0,40.0), bar_mass=10.0,
        agent_radius=40.0, agent_mass=100.0, agent_obs_length=400.0, 
        agent_ray_count=64, agent_actuator_count=16,
        obs_start_angle=0, obs_end_angle=2 * math.pi, 
        act_start_angle=0, act_end_angle=2 * math.pi,
        action_strength=50000.0, friction=0.1, elasticity=0.9,
        food_impulse=10.0, obs_impulse=10.0, bar_impulse=2.0, init_impulse_factor=1.0)->None:

        self.seed(seed)

        self.obss, self.foods, self.bars, self.agent, self.world = \
            None, None, None, None, None
        self.eaten_foods = None
        self.step_reward, self.episod_reward, self.step_count = None, None, None
        self.initial_total_momentum, self.last_total_momentum = None, None
        
        self.food_count, self.obs_count, self.xmax, self.ymax = food_count, obs_count, xmax, ymax
        self.box_size, self.circle_radius, self.box_mass, self.circle_mass = box_size, circle_radius, box_mass, circle_mass
        self.bar_count, self.bar_size, self.bar_mass = bar_count, bar_size, bar_mass
        self.agent_radius, self.agent_mass, self.agent_obs_length = agent_radius, agent_mass, agent_obs_length
        self.agent_ray_count, self.action_strength = agent_ray_count, action_strength
        self.agent_actuator_count = agent_actuator_count
        self.obs_start_angle, self.obs_end_angle = obs_start_angle, obs_end_angle
        self.act_start_angle, self.act_end_angle = act_start_angle, act_end_angle
        self.friction, self.elasticity = friction, elasticity
        self.food_impulse, self.obs_impulse, self.bar_impulse, self.init_impulse_factor = \
            food_impulse, obs_impulse, bar_impulse, init_impulse_factor

        self.action_space = spaces.Discrete(agent_actuator_count+1) # directions clockwise, action 0 is no op
        self.observation_space = spaces.Box(low=0, high=255, shape=(1, agent_ray_count, 3), dtype=np.uint8)

        self._obs_local_pts = list(self._get_pts_on_circle(
            self.observation_space.shape[1], self.agent_obs_length, obs_start_angle, obs_end_angle))
        self._action_pts = list(self._get_pts_on_circle(
            self.action_space.n-1, self.action_strength, act_start_angle, act_end_angle))
        self._agent_mask = 0b1
        self.last_observation = None

        self.renderer = Renderer()

    def reset(self):
        self.obss, self.foods, self.bars = [], [], []
        self.eaten_foods = {}
        self.agent = None
        self.step_reward, self.episod_reward, self.step_count = 0.0, 0.0, 0
        if self.world:
            self.world.end()
        self.world = World(xmax=self.xmax, ymax=self.ymax)
        self.world.create_boundry(collision_type=PodWorldEnv.OBJ_TYPE_WALL)
        self.world.set_collision_callback(PodWorldEnv.OBJ_TYPE_AGENT, PodWorldEnv.OBJ_TYPE_FOOD, 
            self._on_agent_food_collision)
        self.world.set_collision_callback(PodWorldEnv.OBJ_TYPE_AGENT, PodWorldEnv.OBJ_TYPE_ROCK, 
            self._on_agent_rock_collision)
        self.initial_total_momentum, self.last_total_momentum = None, None

        for _ in range(self.obs_count):
            radius = self._get_random_circle_radius()
            obj = CircleBody(mass=self.circle_mass * radius * radius, 
                position=self._get_random_position(clearance=10),
                angle=random.random() * 2 * math.pi, collision_type=PodWorldEnv.OBJ_TYPE_ROCK,
                radius=radius, rgba_color=PodWorldEnv.OBS_COLOR, 
                friction=self.friction, elasticity=self.elasticity)
            self.world.add(obj)
            self.obss.append(obj)            
            obj._apply_gaussian_impulse((radius, radius), obj.mass, obj.mass*self.obs_impulse*self.init_impulse_factor)
        for _ in range(self.food_count):
            box_size = self._get_random_box_size()
            mass = self.box_mass * box_size[0] * box_size[1]
            obj = BoxBody(mass=mass, 
                position=self._get_random_position(clearance=10),
                angle=random.random() * 2 * math.pi, collision_type=PodWorldEnv.OBJ_TYPE_FOOD,
                size=box_size, corner_radius=1.0, rgba_color=PodWorldEnv.AVAIL_FOOD_COLOR, 
                friction=self.friction, elasticity=self.elasticity)
            self.world.add(obj)
            self.foods.append(obj)
            obj._apply_gaussian_impulse(box_size, obj.mass, obj.mass*self.food_impulse*self.init_impulse_factor)
        for i in range(self.bar_count):
            bar_size = self._get_random_bar_size()
            obj = BoxBody(mass=self.bar_mass * bar_size[0] * bar_size[1], 
                position=(self.xmax/(self.bar_count+1)*(i+1), random.randint(0,self.ymax)),
                angle=random.random() * 2 * math.pi, collision_type=PodWorldEnv.OBJ_TYPE_BAR,
                size=bar_size, corner_radius=1.0, rgba_color= (232, 184, 164, 255), 
                friction=self.friction, elasticity=self.elasticity)
            self.world.add(obj)
            self.bars.append(obj)
            obj._apply_gaussian_impulse(bar_size, obj.mass, obj.mass*self.bar_impulse*self.init_impulse_factor)

        # add agent
        self._agent_filter = World.get_filter(self._agent_mask)        
        self.agent = CircleBody(mass=self.agent_mass, 
            position=self._get_random_position(clearance=self.agent_radius),
            angle=random.random() * 2 * math.pi, collision_type=PodWorldEnv.OBJ_TYPE_AGENT,
            radius=self.agent_radius, rgba_color=(224, 23, 33,255), category_mask=self._agent_mask, 
            friction=self.friction, elasticity=self.elasticity)
        self.world.add(self.agent)

    # overriden method
    def seed(self, seed=None):
        np.random.seed(seed)
        random.seed(seed)
        return [seed]

    def step(self, action):
        # perform action
        if action > 0:
            self.agent.apply_impulse(self._action_pts[action-1], (0,0))

        self.world.step(1.0/self.renderer.human_mode_fps)

        if self.step_count % 100 == 0:
            self._update_global_momentum()

        observations = list(self.world.get_observations(self.agent, self._obs_local_pts, self._agent_filter))
        self.last_observation =  np.expand_dims(np.array(observations, dtype=np.uint8), axis=0)

        reward = self.step_reward
        self.step_reward = 0
        self.step_count += 1

        if self.step_count % 100 == 0:
            self._regrow_food()

        return self.last_observation, reward, self.renderer.exited or False, None

    def _on_agent_rock_collision(self, arbiter, space, data)->bool:
        if arbiter.shapes:
            for shape in arbiter.shapes:
                if shape.collision_type == PodWorldEnv.OBJ_TYPE_ROCK:
                    self._add_step_reward(-shape.body.mass * shape.body.velocity.get_length() / 100.0)

        return True
    def _on_agent_food_collision(self, arbiter, space, data)->bool:
        if arbiter.shapes:
            for shape in arbiter.shapes:
                if shape.collision_type == PodWorldEnv.OBJ_TYPE_FOOD and shape not in self.eaten_foods:
                    self._add_step_reward(shape.body.mass)
                    self._set_eaten_status(shape, True)              

        return False        

    def _set_eaten_status(self, shape, is_eaten)->None:
        if is_eaten:
            shape.color = PodWorldEnv.EATEN_FOOD_COLOR
            self.eaten_foods[shape] = self.step_count # record time step when food was eater
        else:
            shape.color = PodWorldEnv.AVAIL_FOOD_COLOR
            self.eaten_foods.pop(shape, None)

    def _regrow_food(self):
        regrown = []
        for shape, eaten_step in self.eaten_foods.items():
            probability = (self.step_count-eaten_step)/shape.body.mass
            if random.random()*probability >= 1:
                regrown.append(shape)
        for shape in regrown:
            self._set_eaten_status(shape, False)

    def _add_step_reward(self, reward:float)->None:
        self.episod_reward += reward
        self.step_reward += reward

    def render(self, mode='human'):
        self.renderer.render(self.world, self.last_observation, self.episod_reward, self.last_total_momentum, mode=mode)

    def close(self):
        self.obss, self.foods = [], []
        if self.world:
            self.world.end()
        self.world = None

    def _get_random_circle_radius(self)->float:
        return max(5.0, abs(random.normalvariate(self.circle_radius, self.circle_radius/2)))

    def _get_random_box_size(self)->float:
        return (max(5.0, abs(random.normalvariate(self.box_size[0], self.box_size[0]/2))),
            max(5.0, abs(random.normalvariate(self.box_size[1], self.box_size[1]/2))))

    def _get_random_bar_size(self)->float:
        return (max(10.0, abs(random.normalvariate(self.bar_size[0], self.bar_size[0]/4))),
            max(10.0, abs(random.normalvariate(self.bar_size[1], self.bar_size[1]/8))))

    def _get_pts_on_circle(self, count:int, radius:float, start_angle:float, end_angle:float)->Iterator[Tuple[float, float]]:
        step = (end_angle-start_angle) / count
        for i in range(count):
            angle = start_angle + step * i
            yield math.cos(angle) * radius, math.sin(angle) * radius

    def _get_random_position(self, clearance)->tuple:
        return random.randint(clearance, self.xmax-clearance), random.randint(clearance, self.ymax-clearance)

    def _update_global_momentum(self):
        self.last_total_momentum = self.world.get_total_momentum()
        if self.initial_total_momentum is None:
            self.initial_total_momentum = self.last_total_momentum
        else:
            decrease = self.initial_total_momentum - self.last_total_momentum
            if decrease > 0:
                print('decreased momentum', decrease)
                obj = random.choice(self.foods)
                obj._apply_gaussian_impulse(obj.size, obj.mass, obj.mass*100)
                obj = random.choice(self.obss)
                obj._apply_gaussian_impulse((obj.radius, obj.radius), obj.mass, obj.mass*100)




import sys
import platform
import numpy as np
import pygame
import pymunk.pygame_util

from ..physics.world import World

class Renderer:
    def __init__(self, viewport_size:tuple=None, human_mode_fps=60, info_height=None)->None:
        self.screen:pygame.Surface = None
        self.human_mode_fps = human_mode_fps
        self.viewport_size = viewport_size
        self.exited = False
        self.info_height = info_height or 100
        self.key_map = {pygame.K_RIGHT:1, pygame.K_UP:0, pygame.K_LEFT:3, pygame.K_DOWN:2}
        self.last_mapped_key:int = None

    def _init_lazy_render(self, world:World):
        if not self.screen:
            if platform.system() == 'Windows':
                # don't scale window size for DPI scaling
                import ctypes
                ctypes.windll.user32.SetProcessDPIAware()            
            pygame.init()
            self.viewport_size = self.viewport_size or (world.xmax, world.ymax)
            self.viewport_size = self.viewport_size[0], self.viewport_size[1] + self.info_height
            self.screen = pygame.display.set_mode(self.viewport_size)
            self.clock = pygame.time.Clock()
            self.font = pygame.font.Font(pygame.font.get_default_font(), 36)

    def render(self, world:World, last_observation:np.ndarray, episod_reward:float, 
        total_momentum:float, mode='human')->np.ndarray:

        self._init_lazy_render(world)
        self._handle_player_events()
        if self.exited:
            return None        
        self._draw_world(world)
        self._draw_info(last_observation, episod_reward, total_momentum)

        ret:np.ndarray = None
        if mode == 'human':
            pygame.display.flip()
            self.clock.tick(self.human_mode_fps)
        else:
            self.clock.tick() # don't introduce additional delay for human vision
            ret = np.flipud(np.rot90(pygame.surfarray.array3d(pygame.display.get_surface())))

        return ret

    def _handle_player_events(self)->None:
        self.last_mapped_key = None

        if self.exited:
            return
        for event in pygame.event.get():
            if event.type == pygame.QUIT or \
                (event.type == pygame.KEYUP and event.key == pygame.K_ESCAPE):

                pygame.quit()
                self.exited = True # sys.exit()   
            elif event.type == pygame.VIDEORESIZE:
                self.viewport_size = event.size
                self.screen = pygame.display.set_mode(self.viewport_size)
                print('Viewport size changed to', self.viewport_size) 
            elif event.type == pygame.KEYDOWN:
                self.last_mapped_key = self.key_map.get(event.key, None)               

    def _draw_info(self, last_observation:np.ndarray, episod_reward:float, total_momentum:float):
        width = self.screen.get_width()
        obs_height = self.info_height / 2.0       
        self.screen.fill(pygame.color.THECOLORS['white'], 
            pygame.rect.Rect(0, 0, width, self.info_height))

        info_text = 'Reward: {:.2f}                                    Global Energy: {:.2f}'.format(episod_reward, total_momentum)
        text_surface = self.font.render(info_text, True, (0, 0, 0))
        self.screen.blit(text_surface, (0,0)) 

        if last_observation is not None:
            obs_width = width / last_observation.shape[1]
            for i in range(last_observation.shape[1]):
                color = tuple(last_observation[0, i])
                self.screen.fill(color, 
                            pygame.rect.Rect(i*obs_width, self.info_height-obs_height, obs_width, obs_height))

    def _draw_world(self, world:World)->None:
        self.screen.fill(world.color)
        options = pymunk.pygame_util.DrawOptions(self.screen)
        options.positive_y_is_up = False
        world.space.debug_draw(options)



import random
from abc import ABCMeta, abstractmethod
import pymunk
from pymunk import Vec2d

class Body(metaclass=ABCMeta):
    def __init__(self)->None:
        self.body:pymunk.Body = None
        self.shape:pymunk.Shape = None

    def apply_impulse(self, impulse:tuple, local_pos:tuple)->None:
        self.body.apply_impulse_at_local_point(impulse, local_pos)
    def apply_force(self, impulse:tuple, local_pos:tuple)->None:
        self.body.apply_force_at_local_point(impulse, local_pos)

    def _apply_gaussian_impulse(self, extent:tuple, magnitude_mean:float, magnitude_stddev:float)->None:
        x, y = extent
        impulse_pos = random.uniform(-x/2, x/2),random.uniform(-y/2, y/2)
        self.apply_impulse(impulse=Body._get_gaussian_impulse(magnitude_mean, magnitude_stddev), 
            local_pos=impulse_pos)

    @staticmethod
    def _get_gaussian_impulse(magnitude_mean:float, magnitude_stddev:float)->tuple:
        direction = Vec2d(random.uniform(-1, 1),random.uniform(-1, 1)).normalized()
        magnitude = random.normalvariate(magnitude_mean, magnitude_stddev)
        return tuple(magnitude * direction)


from typing import Iterator, Callable, Any
import pymunk
import time
from .body import Body

class World:
    def __init__(self, xmax:int, ymax:int, gravityx:float=0.0, gravityy:float=0.0)->None:
        # pymonk has origin on left bottom
        self.xmax, self.ymax = xmax, ymax
        self.space = pymunk.Space()
        self.space.gravity = gravityx, gravityy
        self.last_step_time = time.time()
        self.color = (253, 223, 211, 0)
        self.collision_handlers = []

    def step(self, dt:float=None)->None:
        dt = dt or (time.time() - self.last_step_time)
        self.last_step_time = time.time()
        self.space.step(dt)

    def add(self, body:Body)->None:
        self.space.add(body.shape, body.body)

    def create_boundry(self, width=100, friction:float=0.0, elasticity:float=1.0, collision_type:int=None)->None:
        xmax, ymax = self.xmax, self.ymax
        walls= [pymunk.Segment(self.space.static_body, (-width, -width), (-width, ymax+width), width)
                    ,pymunk.Segment(self.space.static_body, (-width, ymax+width), (xmax+width, ymax+width), width)
                    ,pymunk.Segment(self.space.static_body, (xmax+width, ymax+width), (xmax+width, -width), width)
                    ,pymunk.Segment(self.space.static_body, (-width, -width), (xmax+width, -width), width)
                    ] 
        for s in walls:
            s.friction = friction
            s.elasticity = elasticity
            s.color = (247, 207, 49, 255)
            if collision_type is not None:
                s.collision_type = collision_type

        self.space.add(walls)

    @staticmethod
    def get_filter(value:int)->pymunk.ShapeFilter:
        return pymunk.ShapeFilter(mask=pymunk.ShapeFilter.ALL_MASKS ^ value)

    def get_observations(self, body:Body, local_pts:Iterator[tuple], filter:pymunk.ShapeFilter)->Iterator[tuple]:
        start_pt = body.body.position
        for local_pt in local_pts:
            end_pt = body.body.local_to_world(local_pt)
            result = self.space.segment_query_first(start_pt, end_pt, 0, filter)
            yield result.shape.color if result and result.shape else (0,0,0,0)

    def set_collision_callback(self, collision_type_a:int, collision_type_b:int, 
        callback:Callable[[pymunk.Arbiter, pymunk.Space, Any],bool], at_begin=True)->None:

        ch = self.space.add_collision_handler(collision_type_a, collision_type_b)
        if at_begin:
            ch.begin = callback
        else:
            ch.end = callback
        self.collision_handlers.append(ch)

    def end(self):
        self.collision_handlers.clear()
        self.space = None

    def get_total_momentum(self):
        total_momentum = 0.0
        for body in self.space.bodies:
            if body.body_type == pymunk.Body.DYNAMIC:
                total_momentum += body.mass * body.velocity.get_length()
        return total_momentum


from .body import Body
import pymunk
import numpy as np

class BoxBody(Body):
    def __init__(self, mass:float, position, angle:float, size:tuple, rgba_color:tuple=None, 
        corner_radius:float=0.0, friction:float=0.0, elasticity:float=1.0, category_mask=None,
        collision_type:int=None, inertia=None, inner_radius=0)->None:

        super(BoxBody, self).__init__()
        
        self.size, self.corner_radius, self.mass = size, corner_radius, mass
        self.inertia = inertia or pymunk.moment_for_box(mass, size)
        
        self.body = pymunk.Body(self.mass, self.inertia)
        self.body.position = position
        self.body.angle = angle
        self.shape = pymunk.Poly.create_box(self.body, size, corner_radius)
        self.shape.friction = friction
        self.shape.elasticity = elasticity
        if category_mask:
            self.shape.filter = pymunk.ShapeFilter(categories=category_mask)
        if collision_type is not None:
            self.shape.collision_type = collision_type            
        self.shape.color = rgba_color or tuple(np.random.randint(256, size=3)) + (255,)